@@ -21,7 +21,7 @@ the AsciiDoc format.
 
 package Locale::Po4a::AsciiDoc;
 
-use 5.010;
+use 5.16.0;
 use strict;
 use warnings;
 
@@ -87,7 +87,7 @@ Note that not wrapping the files produced by po4a should not be a
 problem since those files are meant to be processed automatically.
 They should not be regarded as source files anyway.
 
-With this option, po4a will produce better-looking source files, that
+With this option, po4a will produce better-looking AsciiDoc files, but it
 may lead to possibly erroneous formatted outputs.
 
 =item B<noimagetargets>
@@ -107,17 +107,18 @@ Switch parsing rules to compatibility with different tools. Available options ar
 "asciidoc" or "asciidoctor". Asciidoctor has stricter parsing rules, such as
 equality of length of opening and closing block fences.
 
-=item B<yfm_keys>
-
-Comma-separated list of keys to process for translation in the YAML Front Matter
-section. All other keys are skipped. Keys are matched with a case-insensitive
-match. Array values are always translated, unless the B<yfm_skip_array> option
-is provided.
-
 =item B<nolinting>
 
 Disable linting messages. When the source code cannot be fixed for clearer document structure, these messages are useless.
 
+=item B<yfm_keys>
+
+Comma-separated list of keys to process for translation in the YAML Front Matter
+section. All other keys are skipped. Keys are matched with a case-sensitive
+match. If B<yfm_paths> and B<yfm_keys> are used together, values are included if
+they are matched by at least one of the options. Array values are always translated,
+unless the B<yfm_skip_array> option is provided.
+
 =cut
 
 my %yfm_keys = ();
@@ -130,6 +131,19 @@ Do not translate array values in the YAML Front Matter section.
 
 my $yfm_skip_array = 0;
 
+=item B<yfm_paths>
+
+Comma-separated list of hash paths to process for extraction in the YAML
+Front Matter section, all other paths are skipped. Paths are matched with a
+case-sensitive match. If B<yfm_paths> and B<yfm_keys> are used together,
+values are included if they are matched by at least one of the options.
+Arrays values are always returned unless the B<yfm_skip_array> option is
+provided.
+
+=cut
+
+my %yfm_paths = ();
+
 =back
 
 =head1 INLINE CUSTOMIZATION
@@ -205,6 +219,7 @@ sub initialize {
     $self->{options}{'compat'}         = 'asciidoc';
     $self->{options}{'yfm_keys'}       = '';
     $self->{options}{'yfm_skip_array'} = 0;
+    $self->{options}{'yfm_paths'}      = '';
     $self->{options}{'nolinting'}      = 0;
 
     foreach my $opt ( keys %options ) {
@@ -228,8 +243,11 @@ sub initialize {
         $_ =~ s/^\s+|\s+$//g;    # Trim the keys before using them
         $yfm_keys{$_} = 1
     } ( split( ',', $self->{options}{'yfm_keys'} ) );
+    map {
+        $_ =~ s/^\s+|\s+$//g;    # Trim the keys before using them
+        $yfm_paths{$_} = 1
+    } ( split( ',', $self->{options}{'yfm_paths'} ) );
 
-    #        map { print STDERR "key $_\n"; } (keys %yfm_keys);
     $yfm_skip_array = $self->{options}{'yfm_skip_array'};
 
     $self->{translate} = {
@@ -244,6 +262,7 @@ sub initialize {
     $self->register_attributelist('[caption]');
     $self->register_attributelist('[-icons,caption]');
     $self->register_macro('image_[1,alt,title,link]') unless $self->{options}{'noimagetargets'};
+    $self->register_macro('indexterm[1,2,3]') unless $self->{options}{'noimagetargets'};
 
     if ( $self->{options}{'definitions'} ) {
         $self->parse_definition_file( $self->{options}{'definitions'} );
@@ -406,7 +425,8 @@ sub parse {
         my $yamlarray = YAML::Tiny->read_string($yfm)
           || die "Couldn't read YAML Front Matter ($!)\n$yfm\n";
 
-        $self->handle_yaml( $ref, $yamlarray, \%yfm_keys, $yfm_skip_array );
+        $self->handle_yaml( 1, $ref, $yamlarray, \%yfm_keys, $yfm_skip_array, \%yfm_paths );
+        $self->pushline("---\n");
 
         ( $line, $ref ) = $self->shiftline();    # Pass the final '---'
     }
@@ -563,6 +583,8 @@ sub parse {
             do_paragraph( $self, $paragraph, $wrapped_mode );
             $wrapped_mode = 0;
             $paragraph    = "";
+	    $self->pushline( $titlelevel1 . $titlespaces);
+	    $title = $self->translate_indexterms($title);
             my $t = $self->translate(
                 $title,
                 $self->{ref},
@@ -570,7 +592,7 @@ sub parse {
                 "comment" => join( "\n", @comments ),
                 "wrap"    => 0
             );
-            $self->pushline( $titlelevel1 . $titlespaces . $t . $titlelevel2 . "\n" );
+            $self->pushline( $t . $titlelevel2 . "\n" );
             @comments     = ();
             $wrapped_mode = 1;
         } elsif ( ( $line =~ m/^(\/{4,}|\+{4,}|-{4,}|\.{4,}|\*{4,}|_{4,}|={4,}|~{4,})$/ )
@@ -668,9 +690,19 @@ sub parse {
         {
             # Found BlockId
             do_paragraph( $self, $paragraph, $wrapped_mode );
-            $paragraph    = "";
+	    my $block_id = $1;
+	    $paragraph    = "";
             $wrapped_mode = 1;
-            $self->pushline( $line . "\n" );
+	    if ($block_id =~ m/^([^,]+),(.+)$/) {
+		# Found BlockId with a xlabel
+		my $xlabel = $2;
+		$block_id = $1;
+		$self->pushline( "[[$block_id," );
+		do_paragraph( $self, $xlabel, 0);
+		$self->pushline( "]]\n" );
+	    } else {
+		$self->pushline( $line . "\n" );
+	    }
             undef $self->{bullet};
             undef $self->{indent};
         } elsif ( not defined $self->{verbatim}
@@ -699,7 +731,7 @@ sub parse {
             undef $self->{bullet};
             undef $self->{indent};
         } elsif ( not defined $self->{verbatim}
-            and ( $line =~ m/^\[.*\]$/ ) )
+            and ( $line =~ m/^\[[^\]]+\]$/ ) )
         {
             do_paragraph( $self, $paragraph, $wrapped_mode );
             $paragraph = "";
@@ -729,7 +761,7 @@ sub parse {
             undef $self->{bullet};
             undef $self->{indent};
         } elsif ( not defined $self->{verbatim}
-            and ( $line =~ m/^(\s*)([-%~\$[*_+`'#<>[:alnum:]\\"(].*?)((?::::?|;;|\?\?|:-)(?: *\\)?)$/ ) )
+            and ( $line =~ m/^(\s*)([-%~\$[*_+`'#<>[:alnum:]\\"(|\{].*?)((?::::?|;;|\?\?|:-)(?: *\\)?)$/ ) )
         {
             my $indent   = $1;
             my $label    = $2;
@@ -779,13 +811,18 @@ sub parse {
             my $attrname  = $1;
             my $attrsep   = $2;
             my $attrvalue = $3;
-            while ( $attrvalue =~ s/ \+$//s ) {
+	    my $linebreak = "";
+            while ( $attrvalue =~ s/ ([\\+])$//s ) {
+		$linebreak = $1;
+		# add a carriage return at the end of attrvalue if there is none
+		$attrvalue .= "\n" if $attrvalue !~ m/\n$/;
                 ( $line, $ref ) = $self->shiftline();
                 $ref  =~ m/^(.*):[0-9]+$/;
-                $line =~ s/^\s+//;
+                $line =~ s/^\s+|\s+$//;
+		print STDERR "appending attribute with $line" if $debug{parse};
                 $attrvalue .= $line;
             }
-
+	    print STDERR "attr definition: $attrvalue\n" if $debug{parse};
             # Found an Attribute entry
             do_paragraph( $self, $paragraph, $wrapped_mode );
             $paragraph    = "";
@@ -798,10 +835,12 @@ sub parse {
                     $self->{ref},
                     "Attribute :$attrname:",
                     "comment" => join( "\n", @comments ),
-                    "wrap"    => 0
+                    "wrap"    => 1
                 );
+		$t =~ s/\n/ \\\n/g;
                 $self->pushline(":$attrname$attrsep$t\n");
             } else {
+		$attrvalue =~ s/\n/ $linebreak\n/g;
                 $self->pushline(":$attrname$attrsep$attrvalue\n");
             }
             @comments = ();
@@ -839,6 +878,7 @@ sub parse {
                 @comments = ();
             }
         } elsif ( not defined $self->{verbatim}
+		  and ($paragraph =~ m/^\s*$/)
             and ( $line !~ m/^\.\./ )
             and ( $line =~ m/^\.(\S.*)$/ ) )
         {
@@ -860,7 +900,7 @@ sub parse {
             $self->pushline(".$t\n");
             @comments = ();
         } elsif ( not defined $self->{verbatim}
-            and ( $line =~ m/^(\s*)((?:[-*o+\.]+|(?:[0-9]+[.\)])|(?:[a-z][.\)])|\([0-9]+\))\s+)(.*)$/ ) )
+            and ( $line =~ m/^(\s*)((?:(?:[-*o+\.]+(?:\s+\[[ xX\*]\])?)|(?:[0-9]+[.\)])|(?:[a-z][.\)])|\([0-9]+\))\s+)(.*)$/ ) )
         {
             my $indent = $1 || "";
             my $bullet = $2;
@@ -879,6 +919,9 @@ sub parse {
             $paragraph      = $text . "\n";
             $self->{indent} = "";
             $self->{bullet} = $bullet;
+	} elsif ( not defined $self->{verbatim}
+		  and ( $line eq " +") ) {
+	    $paragraph .= $line . "\n";
         } elsif ( ( $line =~ /^\s*$/ ) and ( !defined( $self->{type} ) or ( $self->{type} ne "Table" ) ) ) {
 
             # When not in table, empty lines or lines containing only spaces do break paragraphs
@@ -1018,7 +1061,7 @@ sub parse {
 
             if (   ( $paragraph ne "" && $self->{bullet} && length( $self->{indent} || "" ) == 0 ) )
             {
-		if ( !$self->{options}{'nolinting'} ) {
+		if ( ( !$self->{options}{'nolinting'} ) && ($paragraph !~ m/ \+\n/ ) ) {
 		    # Second line of an item block is not indented. It is unindented
 		    # (and allowed) additional text or a new list item.
 		    warn wrap_mod(
@@ -1040,12 +1083,6 @@ sub parse {
             $paragraph .= $line . "\n";
         }
 
-        # paragraphs starting by a bullet, or numbered
-        # or paragraphs with a line containing many consecutive spaces
-        # (more than 3)
-        # are considered as verbatim paragraphs
-        $wrapped_mode = 0 if ( $paragraph =~ m/^(\*|[0-9]+[.)] )/s
-            or $paragraph =~ m/[ \t][ \t][ \t]/s );
         ( $line, $ref ) = $self->shiftline();
     }
     if ( length $paragraph ) {
@@ -1088,13 +1125,13 @@ sub do_paragraph {
         # - blah         o blah         + blah
         # 1. blah       1) blah       (1) blah
       TEST_BULLET:
-        if ( $paragraph =~ m/^(\s*)((?:[-*o+]|([0-9]+[.\)])|\([0-9]+\))\s+)([^\n]*\n)(.*)$/s ) {
+        if ( $paragraph =~ m/^(\s*)((?:(?:[-*o+](?:\s+\[[ Xx\*]\])?)|([0-9]+[.\)])|\([0-9]+\))\s+)([^\n]*\n)(.*)$/s ) {
             my $para    = $5;
             my $bullet  = $2;
             my $indent1 = $1;
             my $indent2 = "$1" . ( ' ' x length $bullet );
             my $text    = $4;
-            while ( $para !~ m/$indent2(?:[-*o+]|([0-9]+[.\)])|\([0-9]+\))\s+/
+            while ( $para !~ m/$indent2(?:(?:[-*o+](?:\\s+[[ Xx\*]\])?)|([0-9]+[.\)])|\([0-9]+\))\s+/
                 and $para =~ s/^$indent2(\S[^\n]*\n)//s )
             {
                 $text .= $1;
@@ -1133,6 +1170,12 @@ sub do_paragraph {
         $paragraph =~ s/^(.*?)(\n*)$/$1/s;
         $end = $2 || "";
     }
+    if ( defined $self->{bullet} ) {
+        my $bullet  = $self->{bullet};
+        my $indent1 = $self->{indent};
+	$self->pushline($indent1 . $bullet);
+    }
+    $paragraph = $self->translate_indexterms($paragraph);
 
     my $t = $self->translate(
         $paragraph,
@@ -1150,16 +1193,61 @@ sub do_paragraph {
     }
 
     @comments = ();
-    if ( defined $self->{bullet} ) {
+    if ( ( defined $self->{bullet} ) && !($t =~ /\+\n/ ) ) {
         my $bullet  = $self->{bullet};
         my $indent1 = $self->{indent};
         my $indent2 = $indent1 . ( ' ' x length($bullet) );
-        $t =~ s/^/$indent1$bullet/s;
         $t =~ s/\n(.)/\n$indent2$1/sg;
     }
     $self->pushline( $t . $end );
 }
 
+sub translate_indexterms {
+    my ($self, $paragraph) = @_;
+    $paragraph = $self->translate_in_regex($paragraph, qr/\(\(\(([^\)]+)\)\)\)/);
+    return $self->translate_in_regex($paragraph, qr/indexterm:\[([^\]]+)\]/);
+}
+
+sub translate_in_regex {
+    # Detect index entries and translate them separately.
+    # They are moved in front of the paragraph, regardless of their original location,
+    #  but that's consistant with the specification.
+    my ($self, $paragraph, $pattern) = @_;
+    if ( my @indexes = ($paragraph =~ m/$pattern/g ) ) {
+	for my $index (@indexes) {
+	    my @terms = ();
+	    while (
+		$index =~ m/\G(
+       "(?:[^"\\])+"               # quoted term
+       |  (?:[^,\\])+                # unquoted term
+         )(,\s*+)?/gx
+      )
+	    {
+		my $term = $1;
+		if ( $term =~ /^"(.*)"$/ ) {
+		    push @terms, '"' . ($self->translate(
+				  $1,
+				  $self->{ref},
+				  "Index entry",
+				  "wrap" => 1,
+				  "wrapcol" => 0)) . '"';
+		} else {
+		    push @terms, $self->translate(
+				  $term,
+				  $self->{ref},
+				  "Index entry",
+				  "wrap" => 1,
+				  "wrapcol" => 0);
+		}
+	    }
+	    $self->pushline("(((" . join (",", @terms) . ")))");
+
+	}
+    }
+    $paragraph =~ s/$pattern\n?//g;
+    return $paragraph;
+}
+
 sub parse_style {
     my ( $self, $text ) = ( shift, shift );
     $text =~ s/^\[//;
@@ -1321,8 +1409,6 @@ sub unquote_space {
     return $text;
 }
 
-1;
-
 =head1 STATUS OF THIS MODULE
 
 Tested successfully on simple AsciiDoc files.
@@ -1339,7 +1425,11 @@ Tested successfully on simple AsciiDoc files.
  Copyright © 2017 Martin Quinson <mquinson#debian.org>.
 
 This program is free software; you may redistribute it and/or modify it
-under the terms of GPL (see the COPYING file).
+under the terms of GPL v2.0 or later (see the COPYING file).
+
+=cut
+
+1;
 
 __END__
 
